-- Hitfeed rawsets (functions, tables, etc)
-- because it's good to have these in a separate script

-- contains all our hit messages we have to display
rawset(_G, "hf_hitmessages", {})

-- animated icon table
-- {# of frames, tics for each frame}
rawset(_G, "HF_animIcon", {})
HF_animIcon["K_HMBNA"] = {2, 2} -- banana snipe
HF_animIcon["K_HMEGG"] = {2, 2} -- eggbox explode
HF_animIcon["K_HMHYU"] = {2, 2} -- hyudoro
HF_animIcon["K_HMINV"] = {6, 3} -- invincibility
HF_animIcon["K_HMMNE"] = {2, 2} -- mine explode
HF_animIcon["K_HMSNK"] = {2, 2} -- sink
HF_animIcon["K_HMSPB"] = {2, 2} -- spb
HF_animIcon["K_HMSRK"] = {2, 2} -- shrink

-- checks if a player is relevant to the message
rawset(_G, "HF_IsPlayerRelative", function(p)
	if not (netgame) then return false end	-- need to be in a netgame for this
	if not (p and p.valid) then return false end	-- player isnt valid anymore
	
	if (splitscreen)
		-- you can't actually spectate in splitscreen so we can check displayplayers here
		for dp, n in displayplayers.iterate
			if (p == dp) then return true end
		end
	else
		-- since we can spectate, we need to check for consoleplayer
		if (p == consoleplayer) then return true end
	end
	-- we arent a displayplayer
	return false
end)

-- function to get the colour mapping for names
rawset(_G, "HF_ColorToMap", function(c)
	local color_sets = {
	{SKINCOLOR_WHITE, SKINCOLOR_SILVER, SKINCOLOR_SLATE}, -- 1: No colour mapping
	{SKINCOLOR_GREY, SKINCOLOR_NICKEL, SKINCOLOR_BLACK, SKINCOLOR_JET}, -- 2: V_GREYMAP
	{SKINCOLOR_SEPIA, SKINCOLOR_BEIGE, SKINCOLOR_BROWN, SKINCOLOR_LEATHER, SKINCOLOR_RUST}, -- 3: V_BROWNMAP
	{SKINCOLOR_FAIRY, SKINCOLOR_SALMON, SKINCOLOR_PINK, SKINCOLOR_ROSE, SKINCOLOR_BRICK, SKINCOLOR_BUBBLEGUM, SKINCOLOR_LILAC}, -- 4: V_PINKMAP
	{SKINCOLOR_RUBY, SKINCOLOR_RASPBERRY, SKINCOLOR_CHERRY, SKINCOLOR_RED, SKINCOLOR_CRIMSON, SKINCOLOR_MAROON, SKINCOLOR_FLAME, SKINCOLOR_SCARLET, SKINCOLOR_KETCHUP}, -- 5: V_REDMAP
	{SKINCOLOR_DAWN, SKINCOLOR_SUNSET, SKINCOLOR_CREAMSICLE, SKINCOLOR_ORANGE, SKINCOLOR_PUMPKIN, SKINCOLOR_ROSEWOOD, SKINCOLOR_BURGUNDY, SKINCOLOR_TANGERINE}, -- 6: V_ORANGEMAP
	{SKINCOLOR_PEACH, SKINCOLOR_CARAMEL, SKINCOLOR_CREAM}, -- 7: V_PEACHMAP
	{SKINCOLOR_GOLD, SKINCOLOR_ROYAL, SKINCOLOR_BRONZE, SKINCOLOR_COPPER}, -- 8: V_GOLDMAP
	{SKINCOLOR_POPCORN, SKINCOLOR_YELLOW, SKINCOLOR_MUSTARD, SKINCOLOR_OLIVE}, -- 9: V_YELLOWMAP
	{SKINCOLOR_VOMIT, SKINCOLOR_GARDEN, SKINCOLOR_TEA, SKINCOLOR_PISTACHIO}, -- 10: V_TEAMAP
	{SKINCOLOR_LIME, SKINCOLOR_HANDHELD, SKINCOLOR_MOSS, SKINCOLOR_CAMOUFLAGE, SKINCOLOR_ROBOHOOD, SKINCOLOR_MINT, SKINCOLOR_GREEN, SKINCOLOR_PINETREE, SKINCOLOR_EMERALD, SKINCOLOR_SWAMP, SKINCOLOR_DREAM, SKINCOLOR_PLAGUE, SKINCOLOR_ALGAE}, -- 11: V_GREENMAP
	{SKINCOLOR_CARIBBEAN, SKINCOLOR_AQUA, SKINCOLOR_TEAL, SKINCOLOR_CYAN, SKINCOLOR_JAWZ, SKINCOLOR_CERULEAN, SKINCOLOR_NAVY, SKINCOLOR_SAPPHIRE}, -- 12: V_SKYMAP
	{SKINCOLOR_PLATINUM, SKINCOLOR_STEEL}, -- 13: V_STEELMAP
	{SKINCOLOR_PERIWINKLE, SKINCOLOR_BLUE, SKINCOLOR_BLUEBERRY, SKINCOLOR_NOVA}, -- 14: V_BLUEMAP
	{SKINCOLOR_ULTRAVIOLET, SKINCOLOR_PURPLE, SKINCOLOR_FUCHSIA}, -- 15: V_PURPLEMAP
	{SKINCOLOR_PASTEL, SKINCOLOR_MOONSLAM, SKINCOLOR_DUSK, SKINCOLOR_TOXIC, SKINCOLOR_MAUVE, SKINCOLOR_LAVENDER, SKINCOLOR_BYZANTIUM, SKINCOLOR_POMEGRANATE} -- 16: V_LAVENDERMAP
	}
	
	//local maps = {"\x80", "\x86", "\x8e", "\x8d", "\x85", "\x87", "\x8f", "\x8A", "\x82", "\x8b", "\x83", "\x88", "\x8c", "\x84", "\x81", "\x89"}
	local maps = {0, V_GRAYMAP, V_BROWNMAP, V_PINKMAP, V_REDMAP, V_ORANGEMAP, V_PEACHMAP, V_GOLDMAP, V_YELLOWMAP, V_TEAMAP, V_GREENMAP, V_SKYMAP, V_STEELMAP, V_BLUEMAP, V_PURPLEMAP, V_LAVENDERMAP}
	for i = 1, #color_sets
		for j = 1, #color_sets[i]
			local color = color_sets[i][j]
			
			if (color == c)
				return maps[i]
			end
		end
	end
	
	return 0
end)

-- validate each field in the table (this is the saving grace that'll save compatibility with mods!)
rawset(_G, "HF_ValidateFields", function(table)
	if not (table) then return end
	
	-- make sure all the fields are valid in the table
	if not (table[1] and table[2])
	or not (table[1][1] and table[1][2])	-- name and name colour
	or not (table[2][1] and table[2][2])	-- icon and icon colour
		return nil
	end
	
	return table
end)

-- function to send a hitmessage
rawset(_G, "HF_SendHitMessage", function(s, v, icon, color) -- source player, victim player, icon, color (optional)
	-- checks before going through with it
	if not (cv_hf.value) then return end -- hitfeed has to be enabled
	if not (icon) then return end -- if it has no icon we cant do this

	-- now we add the message
	
	-- s and v take table information;
	-- {{name string, name string color}, {icon patch, item patch color}, player}
	-- player is specifically for displayplayer highlighting
	local source = HF_ValidateFields(s)
	local victim = HF_ValidateFields(v)
	
	local hm = {source, victim, icon, TICRATE*5, color or SKINCOLOR_NONE, {1, 1}} -- 6 is for anims
	hf_hitmessages[#hf_hitmessages+1] = hm
end)

-- returns icon for the PlayerSpin hook
rawset(_G, "HF_GetIcon_Spin", function(p, inflictor, source)
	-- easy to return icons through indexing things
	local spinIcons = {
	-- Bananas
	[MT_BANANA] = "K_HMBNA1",
	[MT_BANANA_SHIELD] = "K_HMBNA1",
	-- Orbinauts
	[MT_ORBINAUT] = "K_HMORBN",
	[MT_ORBINAUT_SHIELD] = "K_HMORBN",
	-- Jawz
	[MT_JAWZ] = "K_HMJAWZ",
	[MT_JAWZ_DUD] = "K_HMJAWZ",
	[MT_JAWZ_SHIELD] = "K_HMJAWZ",
	-- Ballhog
	[MT_BALLHOG] = "K_HMBHOG",
	-- SPB
	[MT_SPB] = "K_HMSPB1",
	[MT_SPBEXPLOSION] = "K_HMSPB1",
	-- Lamp oil, ropes, bombs
	[MT_MINEEXPLOSION] = "K_HMMNE1",
	[MT_SSMINE] = "K_HMMNE1",
	[MT_SSMINE_SHIELD] = "K_HMMNE1",
	}

	-- special case: banana snipe
	if not (inflictor.type == MT_BANANA_SHIELD) and (inflictor.type == MT_BANANA)
	and inflictor.health > 1 -- yes this is really how it is....
		return "K_HMBNA"
	end

	-- special case: eggbox
	if inflictor.type == MT_SPBEXPLOSION -- yes eggman boxes are actually spb explosions
	and not inflictor.extravalue1 -- extravalue1 is set on real spb explosions
		return "K_HMEGG1"
	end

	-- no special projectile cases, search up common ones from the table
	if spinIcons[inflictor.type] -- got a hit
		return spinIcons[inflictor.type]
	end

	-- after all that let's check player collision cases (they go last so they dont conflict with others since these are playerstate based)
	if (inflictor.player)
		if (source.player.kartstuff[k_invincibilitytimer]) -- invincibility
			return "K_HMINV"
		end
		if (inflictor.player.kartstuff[k_curshield] == 1) -- thunder shield
			return "K_HMTHNS"
		end
		if (source.player.kartstuff[k_sneakertimer] or source.player.kartstuff[k_pogospring])
			if source.player.kartstuff[k_sneakertimer] -- sneaker steal
				if source.player.kartstuff[k_rocketsneakertimer] -- yes
					return "K_HMRSHE"
				else
					return "K_HMSHOE"
				end
			else -- pogo stomp
				return "K_HMPOGO"
			end
		end
	end

	-- we didn't get anything after all the checks, bummer
	return nil
end)

-- returns the icon for the PlayerSquish hook
rawset(_G, "HF_GetIcon_Squish", function(p, inflictor, source)
	if inflictor.player then
		if source.scale > mapobjectscale then -- grow
			return "K_HMGROW"
		else -- pancakes
			return "K_HMPANK"
		end
	end

	return nil
end)

-- returns the icon for the PlayerExplode hook
rawset(_G, "HF_GetIcon_Explode", function(p, inflictor, source)
	local explodeIcons = {
	[MT_KARMAHITBOX] = "K_HMSPB1",
	[MT_SPBEXPLOSION] = "K_HMSPB",
	-- Lamp oil, ropes, bombs
	[MT_MINEEXPLOSION] = "K_HMMNE",
	[MT_SSMINE] = "K_HMMNE",
	[MT_SSMINE_SHIELD] = "K_HMMNE",
	}
	-- special case: eggbox
	if inflictor.type == MT_SPBEXPLOSION	-- yes eggman boxes are actually spb explosions [...]
	and not inflictor.extravalue1			-- extravalue1 is set on real spb explosions
		return "K_HMEGG"
	end

	if explodeIcons[inflictor.type] -- we got a hit
		return explodeIcons[inflictor.type]
	end

	-- we didn't get anything
	return nil
end)

-- returns var based on splitscreen
rawset(_G, "HF_SplitscreenVar", function(nosplit, split2, split3, split4)
	if (splitscreen)
		if (splitscreen) == 1
			return split2
		else
			return (splitscreen == 3) and split4 or split3
		end
	else
		return nosplit
	end
end)

-- draws backdrop
rawset(_G, "HF_DrawBackdrop", function(v, x, y, length, flags, fadeflags, whitebg)
	local bd, fadeflag = (whitebg) and "HL" or "HT", (whitebg) and fadeflags[2] or fadeflags[1]
	local bd_draw, bd_offset = length, 0
	while (bd_draw)
		local bd_change = 0
		if (bd_draw >= 8)
			if (cv_hf_bd.value) then v.draw(x+bd_offset, y, v.cachePatch("K_"..bd.."MSG8"), flags|fadeflag, v.getColormap(-1, cv_hf_hc.value)) end
			bd_change = 8
		else
			if (cv_hf_bd.value) then v.draw(x+bd_offset, y, v.cachePatch("K_"..bd.."MSG"), flags|fadeflag, v.getColormap(-1, cv_hf_hc.value)) end
			bd_change = 1
		end
		bd_draw, bd_offset = $-bd_change, $+bd_change
	end
	
	return length
end)

-- draws string
rawset(_G, "HF_DrawString", function(v, x, y, str, color, flags)
	-- todo: add long name support
	local c = (cv_hf_cn.value) and HF_ColorToMap(color) or 0
	v.drawString(x, y+2, str, V_ALLOWLOWERCASE|V_6WIDTHSPACE|flags|c, "thin")
	
	return v.stringWidth(str, V_ALLOWLOWERCASE|V_6WIDTHSPACE, "thin")
end)

rawset(_G, "HF_DrawIcon", function(v, x, y, icon, scale, color, flags)
	-- draw icon
	local hx = x+icon.leftoffset
	local height = (icon.height*scale)>>FRACBITS
	local hy = y+(12-height+icon.topoffset)
	v.drawScaled(hx<<FRACBITS, hy<<FRACBITS, scale, icon, flags, v.getColormap(-1, color))
	
	return ((icon.width*scale)>>FRACBITS)
end)

rawset(_G, "HF_MessageWidth", function(v, hm)
	local width = 6 -- start with left border
	
	-- source
	local hf_drawnsource = false
	if (hm[1]) -- check for cached fields (name), source might go nil mid hitmessage	
		-- draw icon
		local hf_drawnicon = false
		if (cv_hf_dm.value != 1)
			width = $+v.cachePatch(hm[1][1][2]).width
				
			hf_drawnicon = true	-- for potential spacing later
		end
			
		-- draw name
		if (cv_hf_dm.value != 2)
			if (hf_drawnicon)	-- spacing
				width = $+1
			end
				
			local str = hm[1][1][1]
			local hf_strtoolong = false
			if (str:len() > cv_hf_cl.value)
				str = str:sub(1, cv_hf_cl.value)
				hf_strtoolong = true
			end
			width = $+v.stringWidth(str, V_ALLOWLOWERCASE|V_6WIDTHSPACE, "thin")
			if (hf_strtoolong)
				width = $+v.stringWidth("..", V_ALLOWLOWERCASE|V_6WIDTHSPACE, "thin")
			end
		end
			
		hf_drawnsource = true	-- for potential spacing later
	end
		
	-- icon
	if (hf_drawnsource)	-- spacing
		width = $+1
	end
	local icon = v.cachePatch(HF_animIcon[hm[3]] and (hm[3]..hm[6][1]) or hm[3])
	width = $+((icon.width*FRACUNIT/2)>>FRACBITS)
		
	-- victim
	if (hm[2]) -- check for cached fields (name), victim might go nil mid hitmessage
		-- initial spacing
		width = $+1
			
		-- draw icon
		local hf_drawnicon = false
		if (cv_hf_dm.value != 1)
			width = $+v.cachePatch(hm[2][1][2]).width
				
			hf_drawnicon = true	-- for potential spacing later
		end
			
		-- draw name
		if (cv_hf_dm.value != 2)
			if (hf_drawnicon)	-- spacing
				width = $+1
			end
				
			local str = hm[2][1][1]
			local hf_strtoolong = false
			if (str:len() > cv_hf_cl.value)
				str = str:sub(1, cv_hf_cl.value)
				hf_strtoolong = true
			end
			width = $+v.stringWidth(str, V_ALLOWLOWERCASE|V_6WIDTHSPACE, "thin")
			if (hf_strtoolong)
				width = $+v.stringWidth("..", V_ALLOWLOWERCASE|V_6WIDTHSPACE, "thin")
			end
		end
	end
		
	-- right border
	width = $+6
	
	return width
end)

-- Overrides the hitfeed locaton, good for mods that change HUD stuff up
-- non splitscreen overrides
if not (hf_overrides) then rawset(_G, "hf_overrides", {0, 0, 0}) end
-- 2p splitscreen overrides
if not (hf_2_overrides) then rawset(_G, "hf_2_overrides", {0, 0, 0}) end
-- 3p/4p splitscreen overrides
if not (hf_34_overrides) then rawset(_G, "hf_34_overrides", {0, 0, 0}) end
rawset(_G, "HF_OverrideLocation", function(xpos, ypos, snapflags, split)
	for p in players.iterate
		if (split) == 2
			if (xpos) then hf_2_overrides[1] = xpos end
			if (ypos) then hf_2_overrides[2] = ypos end
			if (snapflags) then hf_2_overrides[3] = snapflags end
		elseif ((split) == 3 or (split) == 4)
			if (xpos) then hf_34_overrides[1] = xpos end
			if (ypos) then hf_34_overrides[2] = ypos end
			if (snapflags) then hf_34_overrides[3] = snapflags end
		else
			if (xpos) then hf_overrides[1] = xpos end
			if (ypos) then hf_overrides[2] = ypos end
			if (snapflags) then hf_overrides[3] = snapflags end
		end
	end
end)

-- HF_ClearOverrides(split): Clears overrides if they are set
rawset(_G, "HF_ClearOverrides", function(split)
	for p in players.iterate
		if (split) == 2
			hf_2_overrides[1], hf_2_overrides[2], hf_2_overrides[3] = 0, 0, 0
		elseif ((split) == 3 or (split) == 4)
			hf_34_overrides[1], hf_34_overrides[2], hf_34_overrides[3] = 0, 0, 0
		else
			hf_overrides[1], hf_overrides[2], hf_overrides[3] = 0, 0, 0
		end
	end
end)
-- clear overrides on mapload
addHook("MapLoad", function()
	HF_ClearOverrides(1)
	HF_ClearOverrides(2)
	HF_ClearOverrides(3)
end)